/*
  NP_rotate_Cartesian: rotate  North Pole to (gall,galb)

  Copyright (C) 2004-2010 B. Roukema et al. <boud at astro.uni.torun.pl>
  
  This program is free software; you can redistribute it and/or modify
  it under the terms of the GNU General Public License as published by
  the Free Software Foundation; either version 2 of the License, or
  (at your option) any later version.
  
  This program is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
  GNU General Public License for more details.
  
  You should have received a copy of the GNU General Public License
  along with this program; if not, write to the Free Software
  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
  
  See also http://www.gnu.org/licenses/gpl.html
*/  
  
/* possible TODO: switch from matrix right-multiplication to
   left-multiplication */

/* HISTORY: converted from f77 nprotcart.f to C in astromisc-0.1.18.2 */

#include <stdlib.h>
#include "astromisc.h"

/* global variables within this file */
const gsl_rng_type * T;
gsl_rng * r;


/* requires >= C99 compiler */
int printmat(int m, int n, double mat[m][n]){
  int i,j;
  int octave = 1; /* octave style ? */

  if(octave) printf("[");
  for(i=0;i<m;i++){
    for(j=0;j<n;j++){
      printf("%9.2f ",mat[i][j]);
    };
    if(octave){
      if(i<m-1) printf(";");
    } else {
      printf("\n");
    };
  };
  if(octave) printf("]; \n");

  return 0;
}




int NP_rotate_R3(double gal_l,  /*  definition of rotation in DEGREES */
                 double gal_b, 
                 double cart1, /* input position in cartesian coords */
                 double cart2, double cart3, 
                 double *carto1, /* output position in cart coords */
                 double *carto2, double *carto3){

  static double gb_old=-9e9;  /* previous gb, gl if any */
  static double gl_old=-9e9;
  double rotZ[3][3],rotX[3][3]; /* for rotations */
  double rotZinv[3][3]; /* for rotations */
  double rotZX[3][3] = { {0.0,0.0,0.0}, {0.0,0.0,0.0}, {0.0,0.0,0.0}};
  /* full rotation matrix: no need to recalculate if used repeatedly;
     the initialised value should not normally be used, this is just
     a failsafe initialisation.
   */
  static double rotZXZinv[3][3]={ {1.0,0.0,0.0}, {0.0,1.0,0.0}, {0.0,0.0,1.0}}; 

  double cart[3];           /* local: input position in cartesian coords */
  double cartout[3];        /* local: output position in cartesian coords*/
  double phi;  /* temporary angle in radians */

  static const double tol=1e-10;

  int debug=0;
  int ireturn=0;

  cart[0]=cart1;
  cart[1]=cart2;
  cart[2]=cart3;

  if(debug){printf("  NP_rotate_R3: cart= %.2f %.2f %.2f\n",cart[0],cart[1],cart[2]);};

  /*     Is this a new rotation or the previous one?  If it appears
         nearly identical to the previous one, then omit this _if_
         block and use the previously calculated rotation matrix.
   */
  if(fabs(gb_old-gal_b)>tol || fabs(gl_old-gal_l)>tol){
    
    gb_old=gal_b;
    gl_old=gal_l;

    /* rotation around Z-axis, so that 
       [gal_l gal_b] * rotZ 
       is in X-Z plane */
    phi= +gal_l*M_PI_180;  /* astromisc-0.1.8  */

    rotZ[0][0]= cos(phi);
    rotZ[0][1]= sin(phi);
    rotZ[0][2]=0.0;
    rotZ[1][0]= -sin(phi);
    rotZ[1][1]= cos(phi);
    rotZ[1][2]=0.0;
    rotZ[2][0]=0.0;
    rotZ[2][1]=0.0;
    rotZ[2][2]=1.0;

    if(debug){
      printf("rotZ = ");
      printmat(3,3,rotZ);
    };


    /* rotation around Y-axis, so that the North Pole is rotated to
       where (gal_l gal_b) has got to, 
       i.e.
       [0 0 1] * rotZ * rotX =  [gal_l gal_b] * rotZ
    */
    phi= M_PI/2.0 - gal_b*M_PI_180;

    rotX[0][0]= cos(phi);
    rotX[0][1]=0.0;
    rotX[0][2]= sin(phi);
    rotX[2][0]= -sin(phi);
    rotX[2][1]=0.0;
    rotX[2][2]= cos(phi);
    rotX[1][0]=0.0;
    rotX[1][1]=1.0;
    rotX[1][2]=0.0;

    if(debug){
      printf("rotY = ");
      printmat(3,3,rotX);
    };
    /* fortran:         call mult(rotZ,rotX,rotZX,3,3,3,3) */

    /* multiply rotZ (left) by rotX (right) */
    /* gsl_cblas.h: enum CBLAS_ORDER {CblasRowMajor=101, CblasColMajor=102}; */

    cblas_dgemm(102,  /* CblasColMajor = 102 (fortran77 relic) */
                111, 111, /* no transpose */
                3,3,3, 1.0, 
                (double*)rotZ, 3, 
                (double*)rotX, 3,
                0.0, 
                (double*)rotZX, 3);  

    if(debug){
      printf("rotZX = ");
      printmat(3,3,rotZX);
    };

    /* rotate back around Z-axis, so that the North Pole is 
       now shifted to where (gal_l gal_b) was originally, 
       i.e.
       [0 0 1] * rotZ * rotX * rotZinv =  [gal_l gal_b] 
       where
       rotZ_2 is stored in rotZ for convenience
    */

    phi= -gal_l*M_PI_180;  /* astromisc-0.1.8 */

    /* TODO: this can obviously be speeded up if needed */
    rotZinv[0][0]= cos(phi);
    rotZinv[0][1]= sin(phi);
    rotZinv[0][2]=0.0;
    rotZinv[1][0]= -sin(phi);
    rotZinv[1][1]= cos(phi);
    rotZinv[1][2]=0.0;
    rotZinv[2][0]=0.0;
    rotZinv[2][1]=0.0;
    rotZinv[2][2]=1.0;

    if(debug){
      printf("rotZinv = ");
      printmat(3,3,rotZinv);
    };


    /* fortran:         call mult(rotZX,rotZ,rotZXZ,3,3,3,3) */
    cblas_dgemm(102,  /* CblasColMajor = 102 (fortran77 relic) */
                111, 111, /* no transpose */
                3,3,3, 1.0, 
                (double*)rotZX, 3, 
                (double*)rotZinv, 3,
                0.0, 
                (double*)rotZXZinv, 3);  

    if(debug){
      printf("rotZXZinv = ");
      printmat(3,3,rotZXZinv);
    };

  } else { 
    ireturn=1 ; /* 1 means that old value is used */
  } /*   if(fabs(gb_old-gal_b)>tol || fabs(gl_old-gal_l)>tol) */
    

  /* fortran:      call mult(cart,rotZXZ,cartout,1,3,3,3) */
  cblas_dgemm(102,  /* CblasColMajor = 102 (fortran77 relic) */
              111, 111, /* no transpose */
              1,3,3, 1.0, 
              (double*)cart, 1, 
              (double*)rotZXZinv, 3,
              0.0, 
              (double*)cartout, 1);  

  *carto1=cartout[0];
  *carto2=cartout[1];
  *carto3=cartout[2];

  if(debug){printf("  NP_rotate_R3: cartout= %.2f %.2f %.2f\n",
                   cartout[0],cartout[1],cartout[2]);};

  return ireturn;
}



#define DODEC_DIHEDRAL_DEG  116.56505117707799
#define DODEC_DIHEDRAL      2.0344439357957027

/* rotate a dodecahedron from the Gausmann et al. type orientation
   to the old orientation used in gendodec.f in circles */
int rotate_dodec_new2old(
                         double gal_l,  /*  a face centre in DEGREES */
                         double gal_b, 
                         double theta, /* third orientation parameter */
                         double cart1, /* input position in cartesian coords */
                         double cart2, double cart3, 
                         double *carto1, /* output position in cart coords */
                         double *carto2, double *carto3){

  static double gb_old=-9e9;  /* previous gb, gl if any */
  static double gl_old=-9e9;
  static double theta_old=-9e9;
  double rotX[3][3], rotZ[3][3]; /* for rotations */

  /* full rotation matrix: no need to recalculate if used repeatedly;
     the initialised value should not normally be used, this is just
     a failsafe initialisation.
   */
  static double rotXZ[3][3] = { {0.0,0.0,0.0}, {0.0,0.0,0.0}, {0.0,0.0,0.0}};

  double cart[3];           /* local: input position in cartesian coords */
  double cart_theta[3];     /* local: new2old and rotated by theta */
  double cartout[3];        /* local: output position in cartesian coords*/
  double aa;  /* temporary angle in radians */

  static const double tol=1e-10;

  int debug=0;
  int ireturn=0;

  cart[0]=cart1;
  cart[1]=cart2;
  cart[2]=cart3;

  if(fabs(gb_old-gal_b)>tol || fabs(gl_old-gal_l)>tol ||
     fabs(theta_old-theta)>tol){

    gb_old=gal_b;
    gl_old=gal_l;
    theta_old=theta;

    rotX[0][0]=1.0; 
    rotX[0][1]=0.0;
    rotX[0][2]=0.0;
    rotX[1][0]=0.0;
    rotX[1][1]= cos(M_PI_2-DODEC_DIHEDRAL*0.5);
    rotX[1][2]= sin(M_PI_2-DODEC_DIHEDRAL*0.5);
    rotX[2][0]=0.0;
    rotX[2][1]= -sin(M_PI_2-DODEC_DIHEDRAL*0.5);
    rotX[2][2]= cos(M_PI_2-DODEC_DIHEDRAL*0.5);

    if(debug){
      printf("rotX = ");
      printmat(3,3,rotX);
    };


    /* 2pi/10 + third orientation parameter (result in radians)*/
    aa= M_PI*0.1 + M_PI*0.4 - theta * M_PI_180;

    rotZ[0][0]= cos(aa);
    rotZ[0][1]= sin(aa);
    rotZ[0][2]=0.0;
    rotZ[1][0]= -sin(aa);
    rotZ[1][1]= cos(aa);
    rotZ[1][2]= 0.0;
    rotZ[2][0]=0.0;
    rotZ[2][1]=0.0;
    rotZ[2][2]=1.0;

    if(debug){
      printf("rotZ = ");
      printmat(3,3,rotZ);
    };


    cblas_dgemm(102,  /* CblasColMajor = 102 (fortran77 relic) */
                111, 111, /* no transpose */
                3,3,3, 1.0, 
                (double*)rotX, 3, 
                (double*)rotZ, 3,
                0.0, 
                (double*)rotXZ, 3);  

  } else { 
    ireturn=1 ; /* 1 means that old value is used */
  }; /*   if(fabs(gb_old-gal_b)>tol || fabs(gl_old-gal_l)>tol || ...) */
    

  if(debug){
    printf("rotXZ = ");
    printmat(3,3,rotXZ);
  };


  cblas_dgemm(102,  /* CblasColMajor = 102 (fortran77 relic) */
              111, 111, /* no transpose */
              1,3,3, 1.0, 
              (double*)cart, 1, 
              (double*)rotXZ, 3,
              0.0, 
              (double*)cart_theta, 1);  

  /* rotate North Pole down to (gal_l, gal_b) */
  NP_rotate_R3( gal_l, gal_b,
                cart_theta[0],cart_theta[1],cart_theta[2],
                &(cartout[0]),&(cartout[1]),&(cartout[2]) );

  if(debug){printf("  rotate_dodec_new2old: cartout= %.2f %.2f %.2f\n",
                   cartout[0],cartout[1],cartout[2]);};

  *carto1=cartout[0];
  *carto2=cartout[1];
  *carto3=cartout[2];

  return ireturn;
}




