/* 
   astromisc - library with miscellaneous GPL tools for astronomy research
   annulus_limits - asymmetrical min and max angles defining matched annuli

   Copyright (C) 2011 Boud Roukema

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   See also http://www.gnu.org/licenses/gpl.html
*/


/* See Section 3.1 of RBG08 A&A, 2008, 492, 673 (arXiv:0807.4260) */

#include <stdio.h>
#include <strings.h>
#include <math.h>

#define SINCOS_MAX_TOL 1e-11
#define ANNULUS_LIMITS_R_TOL 1e-11

/* OUTPUTS: alpha_(minus|plus) are in radians */
int annulus_limits(int k_curv,  /* INPUTS: normalised sign of curvature */
		   double R_C,  /* curvature radius if k_curv != 0 */
		   double r_SLS, /* radius of SLS */
		   double d_inj, /* twice the injectivity radius */
		   double r_2,   /* maximum cross-separation */
		   double *alpha_minus, /*OUTPUTS: min angle from face centre */
		   double *alpha_plus){ /*OUTPUTS: max angle from face centre */
  double denominator=0.0;
  double cosalpha=0.0;

  int debug=0;

  if(-1==k_curv){

    /* TODO: probably more or less just s/cos/cosh/g && s/sin/sinh/g
       of the +1 cas, but cosh has no upper limit, and maybe some
       plus/minus signs need to be changed. */
    printf("annulus limits k_curve = %d. TODO!\n",k_curv);
    return -1;

  }else if(0==k_curv){

    /* flat equivalent of Sect 3.1 RBG08 */
    denominator = 2.0 * d_inj * r_SLS;
    if(denominator < ANNULUS_LIMITS_R_TOL){
      printf("annulus limits ERROR: denominator= %g too small",denominator);
      printf(", L = %g, r_SLS = %g\n",d_inj,r_SLS);
      return -2;
    };

    /* alpha_minus */
    if(r_2 > 2.0 * r_SLS - d_inj){ /* Eq (6) RBG08 indep of k_curv */
      *alpha_minus = 0.0;
    }else{  /* flat equivalent of Eq (5) RBG08 */
      cosalpha = (d_inj*d_inj + r_SLS*r_SLS - (r_SLS - r_2)*(r_SLS - r_2))/
	denominator;
      if(cosalpha < -1.0 + SINCOS_MAX_TOL){
	*alpha_minus = M_PI;
      }else if(cosalpha > 1.0 - SINCOS_MAX_TOL){
	*alpha_minus = 0.0;
      }else{
	*alpha_minus = acos(cosalpha);
      };
    };
    /* flat equivalent of Eq (5) RBG08 */
    cosalpha = (d_inj*d_inj + r_SLS*r_SLS - (r_SLS + r_2)*(r_SLS + r_2))/
      denominator;
    if(cosalpha < -1.0 + SINCOS_MAX_TOL){
      *alpha_plus = M_PI;
    }else if(cosalpha > 1.0 - SINCOS_MAX_TOL){
      *alpha_plus = 0.0;
    }else{
      *alpha_plus = acos(cosalpha);
    };

  }else if(1==k_curv){

  /* Sect 3.1 RBG08 */
    if(R_C < ANNULUS_LIMITS_R_TOL){
      printf("annulus limits ERROR: R_C= %g too small",R_C);
      printf(", L = %g, r_SLS = %g\n",d_inj,r_SLS);
      return -3;
    };

    denominator = sin(d_inj/R_C) * sin(r_SLS/R_C);
    if(denominator < ANNULUS_LIMITS_R_TOL){
      printf("annulus limits ERROR: denominator= %g too small",denominator);
      printf(", L = %g, r_SLS = %g\n",d_inj,r_SLS);
      return -2;
    };

    /* alpha_minus */
    if(r_2 > 2.0 * r_SLS - d_inj){ /* Eq (6) RBG08  */
      *alpha_minus = 0.0;
    }else{   /* Eq (5) RBG08 */
      cosalpha = ( cos( (r_SLS - r_2)/R_C )
		   -  cos(r_SLS/R_C) * cos(d_inj/R_C) )/
	denominator;
      if(cosalpha < -1.0 + SINCOS_MAX_TOL){
	*alpha_minus = M_PI;
      }else if(cosalpha > 1.0 - SINCOS_MAX_TOL){
	*alpha_minus = 0.0;
      }else{
	*alpha_minus = acos(cosalpha);
      };
    };
    if(debug)printf("A. cosalpha, alpha_plus = %10.5f %10.5f\n",
		    cosalpha, *alpha_plus);
    cosalpha = ( cos( (r_SLS + r_2)/R_C )
		   -  cos(r_SLS/R_C) * cos(d_inj/R_C) )/
      denominator;

    if(debug)printf("B. cosalpha, alpha_plus = %10.5f %10.5f\n",
		    cosalpha, *alpha_plus);
    if(cosalpha < -1.0 + SINCOS_MAX_TOL){
      *alpha_plus = M_PI;
    }else if(cosalpha > 1.0 - SINCOS_MAX_TOL){
      *alpha_plus = 0.0;
    }else{
      *alpha_plus = acos(cosalpha);
    };
    if(debug)printf("C. cosalpha, alpha_plus = %10.5f %10.5f\n",
		    cosalpha, *alpha_plus);

  }else{
    printf("annulus limits ERROR: k_curv= %d invalid.\n",k_curv);
    return -4;
  };
  return 0;
}; 

/* 
   TODO: more automatic testing, e.g. Makefile.am and a script:
   ./astromisc |awk '{print $3,$4}' >> /tmp/a && echo "" >> /tmp/a && ./astromisc |awk '{print $3,$5}' >> /tmp/a && graph -TX -X k=01 -C -W 0.003 -m -4 -S 16 /tmp/a 
*/

int test_annulus_limits(){
  
  int k_curv;  /* INPUTS: normalised sign of curvature */
  double R_C=10000.0;  /* curvature radius if k_curv != 0 */
  double r_SLS=9500.0; /* radius of SLS */
  double d_inj=11700.0; /* twice the injectivity radius */
  double r_2;   /* maximum cross-separation */
  double alpha_minus; /* OUTPUTS: min angle from face centre */
  double alpha_plus; /* OUTPUTS: max angle from face centre */

  int i,n=50;

  for(k_curv=0;k_curv<2;k_curv++){
    for(i=0;i<n;i++){
      r_2 = i/(double)n * r_SLS;
      alpha_minus=-99.9;
      alpha_plus=-99.9;
      annulus_limits( k_curv,  /* INPUTS: normalised sign of curvature */
		      R_C,  /* curvature radius if k_curv != 0 */
		      r_SLS, /* radius of SLS */
		      d_inj, /* twice the injectivity radius */
		      r_2,   /* maximum cross-separation */
		      &alpha_minus, /* OUTPUTS: min angle from face centre */
		      &alpha_plus); /* OUTPUTS: max angle from face centre */
      printf("i,r_2,alpha_mp=  %3d  %10.1f  %10.4f %10.4f\n",
	     i,r_2, alpha_minus, alpha_plus);
    };
    printf("\n");
  };
  return 0;
};

