/* 
   astromisc - library with miscellaneous GPL tools for astronomy research

   Copyright (C) 2004-2008 Boud Roukema

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
*/

#include <stdio.h>
#include <math.h>
#include <gsl/gsl_integration.h>

#include <gsl/gsl_errno.h>
#include <gsl/gsl_interp.h>
#include <gsl/gsl_spline.h>

#include <gsl/gsl_rng.h>

/* gamma functions, factorial etc. */
#include <gsl/gsl_sf_gamma.h>

/* associated legendre functions for spherical harmonics */
#include <gsl/gsl_sf_legendre.h>

#include <gsl/gsl_cblas.h>

#include <stddef.h>
#include <errno.h>
#include <string.h>

#include "config.h"

#include "print_matrices.h"

#define ASTROMISC_WANT_DEBUG 0

/* 
   USER PARAMETERS - BEGIN
*/


/*
   USER PARAMETERS - END
   You probably should not change parameters below unless you're really
   sure you know what you're doing.
*/


/* check generation of  dodecahedral positions on sphere */

#define CHKDOD F77_FUNC(chkdod,CHKDOD)
#ifdef __cplusplus
extern "C"  /* prevent C++ name mangling */
#endif
void CHKDOD( void	   
	     );


/* check generation of  dodecahedral positions on sphere */

#define CVTDOD F77_FUNC(cvtdod,CVTDOD)
#ifdef __cplusplus
extern "C"  /* prevent C++ name mangling */
#endif
void CVTDOD( void	   
	     );


/* fortran wrapper to gsl_rng_env_setup */

#define RAN_OP F77_FUNC_(ran_op,RAN_OP)
#ifdef __cplusplus
extern "C"  /* prevent C++ name mangling */
#endif
void RAN_OP( void	   
	     );


/* fortran wrapper to gsl_rng */
#define RANGSL F77_FUNC(rangsl,RANGSL)
#ifdef __cplusplus
extern "C"  /* prevent C++ name mangling */
#endif
double RANGSL( void	   
	     );


#define RAN_CL F77_FUNC_(ran_cl,RAN_CL)
#ifdef __cplusplus
extern "C"  /* prevent C++ name mangling */
#endif
void RAN_CL( void	   
	     );

/* check sorting of real*4 array */
#define CHKSORT F77_FUNC(chksort,CHKSORT)
#ifdef __cplusplus
extern "C"  /* prevent C++ name mangling */
#endif
void CHKSORT( 
	     );

/* check finding local optima (max or min) */
#define CHKFIND F77_FUNC(chkfind,CHKFIND)
#ifdef __cplusplus
extern "C"  /* prevent C++ name mangling */
#endif
void CHKFIND( 
	     );

/* check pseudorandom number generation functions */
#define CHKRAN F77_FUNC(chkran,CHKRAN)
#ifdef __cplusplus
extern "C"  /* prevent C++ name mangling */
#endif
void CHKRAN( 
	     );



/* sort real*4 array */
#define R4SORT F77_FUNC(r4sort,R4SORT)
#ifdef __cplusplus
extern "C"  /* prevent C++ name mangling */
#endif
void R4SORT( void *array, size_t *nmemb
	     );

int compar_float(const float *x1, const float *x2);


/* hunt_g (gsl) for index in  array */
#define HUNT_G F77_FUNC_(hunt_g,HUNT_G)
#ifdef __cplusplus
extern "C"  /* prevent C++ name mangling */
#endif
void HUNT_G( double x_array[], size_t *nmemb, 
	     double *x_input,
	     size_t *index   /* output */
	     );

/* check hunt_g */
#define CHKHUN F77_FUNC(chkhun,CHKHUN)
#ifdef __cplusplus
extern "C"  /* prevent C++ name mangling */
#endif
void CHKHUN( void
	     );


#define PLMGSL F77_FUNC(plmgsl,PLMGSL)
#ifdef __cplusplus
extern "C"  /* prevent C++ name mangling */
#endif
double PLMGSL(int *l, int *m, double *costheta);


#define FCTGSL F77_FUNC(fctgsl,FCTGSL)
#ifdef __cplusplus
extern "C"  /* prevent C++ name mangling */
#endif
double FCTGSL(int *n);

/* check ylm routine */
#define CHKYLM F77_FUNC(chkylm,CHKYLM)
#ifdef __cplusplus
extern "C"  /* prevent C++ name mangling */
#endif
void CHKYLM( void
	     );


/* generate dodecahedral faces in S^2 in R^3; only 6 characters sign if g77 */
#define GENDODEC F77_FUNC(gendodec,GENDODEC)
#ifdef __cplusplus
extern "C"  /* prevent C++ name mangling */
#endif
void GENDODEC( /* INPUTS */
	      float *gall,  /*  first axis: galactic longitude */
	      float *galb,  /* first axis: galactic latitude */
	      float *theta, /* 3rd parameter for dodecahedron orientation */
	      /* OUTPUTS */
	      float xdod[12], /* 12 centres  x */ 
	      float ydod[12], /* 12 centres  y */ 
	      float zdod[12], /* 12 centres  z */ 
	      /* 12 orthonormal pairs of vectors in R^3 , each
		 orthonormal to the corresponding face centre */
	      float xdodp[2][12], /* x */ 
	      float ydodp[2][12], /* y */ 
	      float zdodp[2][12]  /* z */ 
	     );


/* check holonomy routine */
#define CHKHOL F77_FUNC(chkhol,CHKHOL)
#ifdef __cplusplus
extern "C"  /* prevent C++ name mangling */
#endif
void CHKHOL( void
	     );



/* check Kolmogorov-Smirnov test 2 routine */
#define CHKKS2 F77_FUNC(chkks2,CHKKS2)
#ifdef __cplusplus
extern "C"  /* prevent C++ name mangling */
#endif
void CHKKS2( void
	     );


#ifndef M_PI_180
#define M_PI_180 (M_PI/180.0)
#endif


/* TODO: make these accessible from fortran ? */
int kolmogorov_smirnov_test_2(double *x,    /* inputs */
			      int   n_x,
			      double *y, 
			      int   n_y,
			      char *alt,      /* inputs: TODO */
			      double *pval,   /* outputs: TODO */
			      double *ks,     /* outputs: DONE */
			      double *d_ks    /* outputs: DONE */
			      );

int NP_rotate_R3(double gal_l,  /*  definition of rotation in DEGREES */
                 double gal_b, 
                 double cart1, /* input position in cartesian coords */
                 double cart2, double cart3, 
                 double *carto1, /* output position in cart coords */
                 double *carto2, double *carto3);


int rotate_dodec_new2old(
                         double gal_l,  /*  a face centre in DEGREES */
                         double gal_b, 
                         double theta, /* third orientation parameter */
                         double cart1, /* input position in cartesian coords */
                         double cart2, double cart3, 
                         double *carto1, /* output position in cart coords */
                         double *carto2, double *carto3);


int test_NP_rotate_R3(void);

int test_rotate_dodec_new2old(void);

int test_rotate_about_axis3D(void);


int test_orth_axes(void);

int test_get_nearby_point(void);

int test_c_angle_R4(void);

int test_c_hyperangle_R4(void);

int test_separation_3D(void);

int test_genholon_flat(void);

int test_sim_galaxy_second_point(void);

int test_parse_noempty_strtox(void);
