/* 
   astromisc - library with miscellaneous GPL tools for astronomy research
   c_angle_R4 - angle between two vectors in R4, possibly using only some dims

   Copyright (C) 2011 Boud Roukema

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   See also http://www.gnu.org/licenses/gpl.html
*/


#include <stdio.h>
#include <strings.h>
#include <math.h>
#include "holonomies.h"

/* return value in radians */
double c_angle_R4(struct xyzw_s vec_a, /* INPUT vectors */
		  struct xyzw_s vec_b, 
		  int n_dim            /* use just 3 or all 4 dims? */
		  ){
  /* return value */
  double angle;
  /* local */
  double r1sq, r1, r2sq, r2, dot, costh;

  if(3==n_dim){
    r1sq = vec_a.x*vec_a.x + vec_a.y*vec_a.y + vec_a.z*vec_a.z ;
    r1=sqrt(r1sq);
    if(r1 <= SMALL_R_TOL)
      printf("c_angle_R4 ERROR: r1 = %g\n",r1);

    r2sq = vec_b.x*vec_b.x + vec_b.y*vec_b.y + vec_b.z*vec_b.z ;
    r2=sqrt(r2sq);
    if(r2 <= SMALL_R_TOL)
      printf("c_angle_R4 ERROR: r2 = %g\n",r2);
    
    dot=  vec_a.x*vec_b.x + vec_a.y*vec_b.y + vec_a.z*vec_b.z ;
    costh = dot/r1/r2;
    if(costh < -1.0 + SINCOS_MAX_TOL) costh= -1.0 + SINCOS_MAX_TOL;
    if(costh > 1.0 - SINCOS_MAX_TOL) costh= 1.0 - SINCOS_MAX_TOL;

    angle= acos(costh);
    return angle;
  }else  if(4==n_dim){
    r1sq = vec_a.x*vec_a.x + vec_a.y*vec_a.y + vec_a.z*vec_a.z + vec_a.w*vec_a.w;
    r1=sqrt(r1sq);
    if(r1 <= SMALL_R_TOL)
      printf("c_angle_R4 ERROR: r1 = %g\n",r1);

    r2sq = vec_b.x*vec_b.x + vec_b.y*vec_b.y + vec_b.z*vec_b.z + vec_b.w*vec_b.w;
    r2=sqrt(r2sq);
    if(r2 <= SMALL_R_TOL)
      printf("c_angle_R4 ERROR: r2 = %g\n",r2);
    
    dot=  vec_a.x*vec_b.x + vec_a.y*vec_b.y + vec_a.z*vec_b.z + vec_a.w*vec_b.w;
    costh = dot/r1/r2;
    if(costh < -1.0 + SINCOS_MAX_TOL) costh= -1.0 + SINCOS_MAX_TOL;
    if(costh > 1.0 - SINCOS_MAX_TOL) costh= 1.0 - SINCOS_MAX_TOL;

    angle= acos(costh);
    return angle;
  }else{
    printf("c_angle_R4: n_dim =%d not programmed. Sorry.\n",n_dim);
    return 1;
  };
};


int test_c_angle_R4(void){

  struct xyzw_s vec_a; /* INPUT vectors */
  struct xyzw_s vec_b; 
  double a;

  vec_a.x = 1.0/sqrt(2.0);
  vec_a.y = 1.0/sqrt(2.0);
  vec_a.z = 0.0;

  vec_b.x = 1.0/sqrt(2.0);
  vec_b.y = -1.0/sqrt(2.0);
  vec_b.z = 0.0;

  /*
  vec_a.x = 1.0/sqrt(2.0);
  vec_a.y = 0.0;
  vec_a.z = 1.0/sqrt(2.0);

  vec_b.x = 1.0/sqrt(2.0);
  vec_b.y = 0.0;
  vec_b.z = -1.0/sqrt(2.0);
  */

  printf("angle = %20.15f\n",
	 c_angle_R4(vec_a,vec_b, 3));
  vec_a.x *= 99.9;
  vec_a.y *= 99.9;
  vec_a.z *= 99.9;
  printf("same angle = %20.15f\n",
	 c_angle_R4(vec_a,vec_b, 3));

  vec_b.x = 1.0/sqrt(2.0);
  vec_b.y = 1.0/sqrt(2.0);
  vec_b.z = 1.0/sqrt(3.0); /* 1e-5 */ /* 500.0 */

  a = c_angle_R4(vec_a,vec_b, 3);
  printf("30 deg angle = %20.15f, *180/M_PI = %20.15f\n",
	 a, a*180.0/M_PI);

  return 0;
};
