/* 
   astromisc - library with miscellaneous GPL tools for astronomy research
   c_hyperangle_R4 - hyperangle between two vectors in R4, possibly using only some dims

   Copyright (C) 2012 Boud Roukema

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   See also http://www.gnu.org/licenses/gpl.html
*/


#include <stdio.h>
#include <strings.h>
#include <math.h>
#include "holonomies.h"

/* 
   2D: z is the symmetry axis of the 2-hyperboloid in R3
   3D: w is the symmetry axis of the 3-hyperboloid in R4
 */
double c_hyperangle_R4(struct xyzw_s vec_a, /* INPUT vectors */
		       struct xyzw_s vec_b, 
		       int n_dim            /* use just 3 or all 4 dims? */
		       ){
  /* return value */
  double hyperangle;
  /* local */
  double r1sq, r1, r2sq, r2, dot, costh;

  if(3==n_dim){
    r1sq = - vec_a.x*vec_a.x - vec_a.y*vec_a.y  + vec_a.z*vec_a.z ;
    if(r1sq <= SMALL_R_TOL)
      printf("c_hyperangle_R4 ERROR: r1sq = %g\n",r1sq);
    r1=sqrt(r1sq);

    r2sq = - vec_b.x*vec_b.x - vec_b.y*vec_b.y  + vec_b.z*vec_b.z ;
    if(r2sq <= SMALL_R_TOL)
      printf("c_hyperangle_R4 ERROR: r2sq = %g\n",r2sq);
    r2=sqrt(r2sq);
    
    dot= - vec_a.x*vec_b.x - vec_a.y*vec_b.y  + vec_a.z*vec_b.z ;
    costh = dot/r1/r2;

    hyperangle= acosh(costh);
    return hyperangle;
  }else  if(4==n_dim){
    r1sq = - vec_a.x*vec_a.x - vec_a.y*vec_a.y - vec_a.z*vec_a.z + vec_a.w*vec_a.w;
    if(r1sq <= SMALL_R_TOL)
      printf("c_hyperangle_R4 ERROR: r1sq = %g\n",r1sq);
    r1=sqrt(r1sq);

    r2sq = - vec_b.x*vec_b.x - vec_b.y*vec_b.y - vec_b.z*vec_b.z + vec_b.w*vec_b.w;
    if(r2sq <= SMALL_R_TOL)
      printf("c_hyperangle_R4 ERROR: r2sq = %g\n",r2sq);
    r2=sqrt(r2sq);
    
    dot=  - vec_a.x*vec_b.x - vec_a.y*vec_b.y - vec_a.z*vec_b.z + vec_a.w*vec_b.w;
    costh = dot/r1/r2;

    hyperangle= acosh(costh);
    return hyperangle;
  }else{
    printf("c_hyperangle_R4: n_dim =%d not programmed. Sorry.\n",n_dim);
    return 1;
  };
};


int test_c_hyperangle_R4(void){

  struct xyzw_s vec_a; /* INPUT vectors */
  struct xyzw_s vec_b; 
  double a;

  vec_a.x = 3.0 / sqrt(2.0);
  vec_a.y = 3.0 / sqrt(2.0);
  vec_a.z = 5.0;

  vec_b.x = -3.0 / sqrt(2.0);
  vec_b.y = -3.0 / sqrt(2.0);
  vec_b.z = 5.0;

  /*
  vec_a.x = 1.0/sqrt(2.0);
  vec_a.y = 0.0;
  vec_a.z = 1.0/sqrt(2.0);

  vec_b.x = 1.0/sqrt(2.0);
  vec_b.y = 0.0;
  vec_b.z = -1.0/sqrt(2.0);
  */

  printf("hyperangle = %20.15f; cf 2.0* atanh(0.60) = %20.15f\n",
	 c_hyperangle_R4(vec_a,vec_b, 3), 2.0* atanh(0.60));
  vec_a.x *= 99.9;
  vec_a.y *= 99.9;
  vec_a.z *= 99.9;
  printf("same hyperangle = %20.15f\n",
	 c_hyperangle_R4(vec_a,vec_b, 3));

  vec_b.x = 0.0;
  vec_b.y = 0.0;
  vec_b.z = 4.0; /* 1e-5 */ /* 500.0 */

  a = c_hyperangle_R4(vec_a,vec_b, 3);
  printf("hyperangle = %20.15f; cf  atanh(0.60) = %20.15f\n",
	 a,  atanh(0.60));

  return 0;
};
