/* 
   astromisc - library with miscellaneous GPL tools for astronomy research
   c_xyz_rlb - convert 3D XYZ position to radius, latitude, longitude in rad

   Copyright (C) 2004-2011 Boud Roukema

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   See also http://www.gnu.org/licenses/gpl.html
*/

#include <math.h>

/* this group are only needed for testing */
#include <stdio.h>
#include <strings.h>
#include <gsl/gsl_rng.h>


#define SINCOS_MAX_TOL 1e-11
#define GLONGLAT_UNDEF 999.9
/* SMALL_R_TOL depends on physical units, in principle 1e-50 
   should be no problem if all distance units are consistent */
#define SMALL_R_TOL  1e-100


/* convert 3D Euclidean XYZ position to 
   radial distance, latitude and longitude coords 
   in radians
   * Warning: longitude first (as in fortran version)
*/
int c_xyz_rlb(double x_in,  /* inputs */
	      double y_in,
	      double z_in, 
	      double *radius, /* outputs */
	      double *glong,
	      double *glat){

  double z_unit;
  double cos_glat, cos_glong;

  *radius=x_in*x_in + y_in*y_in + z_in*z_in;
  if(*radius<SMALL_R_TOL){
    *glong = GLONGLAT_UNDEF;
    *glat =  GLONGLAT_UNDEF;
    return 1;
  }else{
    *radius = sqrt(*radius);
    
    z_unit = z_in / (*radius);
    if(z_unit > 1.0-SINCOS_MAX_TOL){
      z_unit = 1.0-SINCOS_MAX_TOL;
    }else if(z_unit < -1.0+SINCOS_MAX_TOL){
      z_unit = -1.0+SINCOS_MAX_TOL;
    };

    *glat = asin(z_unit);

    cos_glat = cos(*glat);

    if(cos_glat < SINCOS_MAX_TOL){
      *glong = GLONGLAT_UNDEF; /* since close to NP or SP */
    }else{
      cos_glong = x_in / (*radius) / cos_glat;
      if(cos_glong > 1.0-SINCOS_MAX_TOL){
	cos_glong = 1.0-SINCOS_MAX_TOL;
      }else if(cos_glong < -1.0+SINCOS_MAX_TOL){
	cos_glong = -1.0+SINCOS_MAX_TOL;
      };

      *glong = acos(cos_glong);
      if(y_in < 0.0){
	*glong = 2.0*M_PI - *glong;
      };
    }; /* if(cos_glat < SINCOS_MAX_TOL) else ... */
  }; /*   if(*radius<SMALL_R_TOL) else ... */
  return 0;
};

