/* 
   astromisc - library with miscellaneous GPL tools for astronomy research
   get_orth_axes, get_nearby_point - get point on S^2 at nearby 3D distance

   Copyright (C) 2004-2011 Boud Roukema

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   See also http://www.gnu.org/licenses/gpl.html
*/


#include <stdio.h>
#include <strings.h>
#include <math.h>
#include <gsl/gsl_rng.h>
#include "holonomies.h"

#define SINCOS_MAX_TOL 1e-11
/* SMALL_R_TOL depends on physical units, in principle 1e-50 
   should be no problem if all distance units are consistent */
#define SMALL_R_TOL_PHYS  1e-50
#define NEARBY_POINT_GSL_SEED_OFFSET 1988438083
#define GLONGLAT_UNDEF 999.9

/* Given a sky position (glong_deg,glat_deg) calculate the Euclidean
   XYZ (unit sphere) position of that same position, and two
   orthonormal axes. */
int get_orth_axes(double glong,      /* INPUTS: sky coords gal long radians */
		  double glat,       /* sky coords gal lat radians */
		  double vec_lb[3],      /* OUTPUTS: sky position in XYZ */
		  double ortho_a[3],     /* one orthonormal axis XYZ */
		  double ortho_b[3]){     /* final orthonormal axis XYZ */

  /* local */
  double cb, sb, cl, sl;  /* cos(b), sin(b), cos(l), sin(l) */

  cb = cos(glat);
  sb = sin(glat);
  cl = cos(glong);
  sl = sin(glong);

  vec_lb[0] = cb*cl;
  vec_lb[1] = cb*sl;
  vec_lb[2] = sb;

  ortho_a[0] = -sl;
  ortho_a[1] = cl;
  ortho_a[2] = 0.0;
  
  ortho_b[0] = -sb*cl;
  ortho_b[1] = -sb*sl;
  ortho_b[2] = cb;
  
  return 0;
};


/* R^3, R^4 functions defined locally to be sure what is used here */
double  dot_R3(double a[3], double b[3]){
  double dot;
  dot = a[0]*b[0] + a[1]*b[1] + a[2]*b[2];
  return dot;
};

double  dist_R3(double a[3], double b[3]){
  double xx,yy,zz;
  xx=a[0]-b[0];
  yy=a[1]-b[1];
  zz=a[2]-b[2];
  return sqrt(xx*xx+yy*yy+zz*zz);
};

/* R^4 functions defined locally to be sure what is used here */
double  length_R4_xyzw(struct xyzw_s vector4){
  double norm;
  norm = vector4.x * vector4.x * +
    vector4.y * vector4.y * +    
    vector4.z * vector4.z * +    
    vector4.w * vector4.w ;
  if(norm > 0.0){
    norm = sqrt(norm);
  }else{
    norm = 0.0;
  };
  return norm;
};


/* Given a rotational axis sky position (glong_axis,glat_axis), a sky
   position to be rotated (glong_in,glat_in) and a rotation angle
   (radians), rotate the g*_in position by the rotation angle. */
   
int rotate_about_axis3D(double glong_axis, /* INPUTS: axis gal long radians */
			double glat_axis,       /* sky coords gal lat radians */
			double theta,          /* amount to rotate radians */
			double *glong_in,       /* position to rotate g long */
			double *glat_in){       /*           gal lat radians */

  /* local */
  double vec_axis[3];      /* axis position in XYZ */
  double ortho_a[3];     /* one orthonormal axis XYZ */
  double ortho_b[3];     /* another orthonormal axis XYZ */

  double vec_in[3];      /* input position in XYZ */
  double dummy[3];
  double vec_in_axis_basis[3]; /* vec_in expressed in rotation axis basis */
  double vec_rotated_axis_basis[3]; /* after rotation, in axis basis */
  double vec_rotated_astro_basis[3]; /* after rotation, in astronomical basis */
  double radius;  /* needed for c_xyz_rlb() */

  int i;

  /* get orthonormal basis for axis about which to rotate */
  get_orth_axes(glong_axis,glat_axis, vec_axis, ortho_a, ortho_b);
  /* get vector version of input position */
  get_orth_axes(*glong_in,*glat_in, vec_in, dummy, dummy);

  /* express the vector in the basis defined by the rotation axis */
  vec_in_axis_basis[0] = dot_R3(vec_in, vec_axis);
  vec_in_axis_basis[1] = dot_R3(vec_in, ortho_a);
  vec_in_axis_basis[2] = dot_R3(vec_in, ortho_b);

  /* rotate it */
  vec_rotated_axis_basis[0] =  vec_in_axis_basis[0]; /* on axis: no effect */
  vec_rotated_axis_basis[1] =  cos(theta)*vec_in_axis_basis[1] -
    sin(theta) * vec_in_axis_basis[2];
  vec_rotated_axis_basis[2] =  sin(theta)*vec_in_axis_basis[1] +
    cos(theta) * vec_in_axis_basis[2];
  
  /* express it in the astronomical basis set */
  for(i=0;i<3;i++){
    vec_rotated_astro_basis[i] = 
      vec_rotated_axis_basis[0] * vec_axis[i] + 
      vec_rotated_axis_basis[1] * ortho_a[i] + 
      vec_rotated_axis_basis[2] * ortho_b[i];
  };

  c_xyz_rlb(vec_rotated_astro_basis[0],
	    vec_rotated_astro_basis[1], 
	    vec_rotated_astro_basis[2],
	    &radius, glong_in, glat_in);

  return 0;
};



/* Given a 4D position, the radius to the surface of last scattering,
   and a spatial geodesic distance, randomly choose another point on
   the SLS at that distance, from a uniform distribution along the
   circle of possible points. Return value: 0 = OK, -1 = no intersection.
   TODO: Add the hyperbolic case.  */
int get_nearby_point(struct xyzw_s posn_in, /* INPUTS: spatial position */
		     double distance,     /* distance to new point */
		     int    k_curv,         /* sign of curvature  */
		     double R_C,  /* curvature radius if k_curv != 0 */
		     double rSLS,           /* radius of SLS */
		     int    reset_gsl_rng,   /* free up gsl rng after usage? */
		     double *glong_deg_out,  /* OUTPUTS: sky coords */
		     double *glat_deg_out
		  ){

  double vec_lb_in[3];      /* sky position in XYZ */
  double ortho_a[3];     /* one orthonormal axis XYZ */
  double ortho_b[3];     /* final orthonormal axis XYZ */

  double cos_theta_distance;
  double theta_distance;  /* distance converted to an angle in
			     radians on the SLS */

  /* declarations  struct xyzw_s posn_in_minus_origin_4Dchord
     for GSL pseudo-random number generator */
  static int gsl_initialised = 0;
  static const gsl_rng_type *T;
  static gsl_rng *r;

  double phi; /* random angle around vector vec_lb_in */
  double cth,sth, cph,sph; /* cos and sin of theta and phi */
  double vec_lb_out[3]; /* output position as a unit vector */
  double posn_in_radius, radius;
  double posn_in_4D_angle=-9e9; /* used for non-flat case(s) */
  struct xyzw_s origin_xyzw = {0.0, 0.0, 0.0, 1.0};
  double glong,glat;
  int i;

  int debug=0; 
  int want_verbose=0; /* TODO: could feed this from calling program */

  /* get orthonormal position of input point and two orthonormal
     axes */
  c_xyz_rlb(posn_in.x, posn_in.y, posn_in.z, 
	    &posn_in_radius, &glong, &glat);
  if(debug)printf("\nposn_in_radius, glong glat = %f, %f %f rad %f %f deg\n",
		  posn_in_radius, glong, glat,
		  glong*180.0/M_PI, glat*180.0/M_PI
		  );
  get_orth_axes(glong,glat, 
	       vec_lb_in, ortho_a, ortho_b);
  if(debug){
    printf("posn_in = %.2f  %.2f  %.2f  %.2f,  rSLS, distance = %.2f %.2f\n",
	   posn_in.x,posn_in.y,posn_in.z,posn_in.w, 
	   rSLS, distance);
    printf("vec_lb_in = %.2f  %.2f  %.2f,  norm = %.4f\n",
	   vec_lb_in[0],vec_lb_in[1],vec_lb_in[2],
	   dot_R3(vec_lb_in,vec_lb_in) );
    printf("ortho_a = %.2f  %.2f  %.2f,  norm = %.4f\n",
	   ortho_a[0],ortho_a[1],ortho_a[2],
	   dot_R3(ortho_a,ortho_a) );
    printf("ortho_b = %.2f  %.2f  %.2f,  norm = %.4f\n",
	   ortho_b[0],ortho_b[1],ortho_b[2],
	   dot_R3(ortho_b,ortho_b) );
  };

  if(1==k_curv){
    origin_xyzw.w  *= R_C; /* put the w coordinate of the observing
			      origin at R_C */
    posn_in_4D_angle = c_angle_R4(posn_in, origin_xyzw, 4);
    if(debug){
      printf("get_nearby_point: posn_in_4D_angle = %.5f should = ",posn_in_4D_angle); 
      printf("asin(posn_in_radius/R_C) = %.5f\n", asin(posn_in_radius/R_C));
      printf("get_nearby_point: TEST_0 diff should be small: %10g\n",
	     posn_in_4D_angle - asin(posn_in_radius/R_C)); 
    }

    posn_in_radius = R_C * posn_in_4D_angle; /* earlier value is an R^3
					      projected value */
  }

  if(k_curv < 0){
    printf("get_nearby_point ERROR: k_curv = %d not yet programmed.\n",
	   k_curv);
    exit(1);
  };
  if(k_curv > 1){
    printf("get_nearby_point ERROR: k_curv = %d  is invalid.\n",
	   k_curv);
    exit(1);
  };

  /* triangle inequality in flat space: is there an intersection with
     the SLS? 
     * this should also work in spherical space
  */
  if(debug){
    printf("rSLS = %f distance = %f posn_in_radius = %f\n",
	   rSLS , distance , posn_in_radius );
  };
  if( !( rSLS - distance < posn_in_radius &&
	 posn_in_radius < rSLS + distance ) ){
    if(want_verbose){
      printf("get_nearby_point: no intersection\n");
    };
    *glong_deg_out = GLONGLAT_UNDEF;
    *glat_deg_out = GLONGLAT_UNDEF;
    return -1;
  };

  if(0==k_curv){
    /* cosine rule in flat space */
    cos_theta_distance = rSLS*rSLS + posn_in_radius*posn_in_radius 
      - distance*distance ;
    if(rSLS < SMALL_R_TOL_PHYS || posn_in_radius < SMALL_R_TOL_PHYS){
      printf("get_nearby_point Warning: rSLS = %10g ",rSLS);
      printf(" or posn_in_radius = %10g too small.\n",posn_in_radius);
      cos_theta_distance /= 2.0*SMALL_R_TOL_PHYS*SMALL_R_TOL_PHYS;  
    }else{
      cos_theta_distance /= 2.0*rSLS*posn_in_radius;    
    };
  }else if(1==k_curv){
    /* cosine rule in spherical space */
    if(rSLS < SMALL_R_TOL_PHYS || posn_in_radius < SMALL_R_TOL_PHYS){
      printf("get_nearby_point WARNING: rSLS = %10g ",rSLS);
      printf(" or posn_in_radius = %10g too small.\n",posn_in_radius);
      cos_theta_distance = 0.0;
    }else{
      cos_theta_distance = ( cos(distance/R_C) 
			     - cos(rSLS/R_C) * cos(posn_in_4D_angle) ) /
	(sin(rSLS/R_C) * sin(posn_in_4D_angle));
    };
  };
    
  if(debug)printf("cos_theta_distance = %.5g, ",cos_theta_distance);
  if(cos_theta_distance < -1.0 + SINCOS_MAX_TOL){
    cos_theta_distance = -1.0 + SINCOS_MAX_TOL;
  }else if(cos_theta_distance > 1.0 - SINCOS_MAX_TOL){
    cos_theta_distance = 1.0 - SINCOS_MAX_TOL;
  };
  if(debug)printf(" %.5f\n",cos_theta_distance);
    
  theta_distance = acos(cos_theta_distance);
  if(debug)printf("theta_distance = %.5f, ",theta_distance);

  if(!gsl_initialised){
    /* set up the generator */
    gsl_rng_env_setup();
    T = gsl_rng_default;
    r = gsl_rng_alloc (T);
    gsl_rng_default_seed += NEARBY_POINT_GSL_SEED_OFFSET;
    gsl_initialised = 1;
  };

  phi= 2.0*M_PI * gsl_rng_uniform (r); /* random angle in orthogonal 2-plane */
  cth=cos(theta_distance);
  sth=sin(theta_distance);
  cph=cos(phi);
  sph=sin(phi);
  if(debug)printf("cth= %.4f sth = %.4f  mod^2 = %.4f\n",
		  cth,sth, cth*cth+sth*sth);

  for(i=0;i<3;i++){
    vec_lb_out[i] = 
      cth* vec_lb_in[i] +
      sth* (cph*ortho_a[i] + sph*ortho_b[i]);
  };

  c_xyz_rlb(vec_lb_out[0], vec_lb_out[1], vec_lb_out[2], 
	    &radius, &glong, &glat);
  *glong_deg_out = glong*180.0/M_PI;
  *glat_deg_out  = glat*180.0/M_PI;

  if(reset_gsl_rng){
    gsl_rng_free (r);
    gsl_initialised = 0;
  };

  return 0;
};

/*  tests  */



int test_orth_axes(void){
  
  /* declarations for GSL pseudo-random number generator */
  const gsl_rng_type *T;
  gsl_rng *r;

  double b,l;
  double v[3], ortho_a[3], ortho_b[3];
  int i, n= 10;

  /* set up the generator */
  gsl_rng_env_setup();
  T = gsl_rng_default;
  r = gsl_rng_alloc (T);

  for(i=0;i<n;i++){
    b = M_PI* gsl_rng_uniform (r) - M_PI_2;
    l = 2.0*M_PI * gsl_rng_uniform (r);
    get_orth_axes(l, b, v, ortho_a, ortho_b);
    printf("i = %d, l = %.2f, b = %.2f \n",i,l,b);
    /* check orthonormality */
    printf("%10.6f ", dot_R3(v,v));
    printf("%10.6f ", dot_R3(v,ortho_a));
    printf("%10.6f \n", dot_R3(v,ortho_b));

    printf("%10.6f ", dot_R3(ortho_a,v));
    printf("%10.6f ", dot_R3(ortho_a,ortho_a));
    printf("%10.6f \n", dot_R3(ortho_a,ortho_b));

    printf("%10.6f ", dot_R3(ortho_b,v));
    printf("%10.6f ", dot_R3(ortho_b,ortho_a));
    printf("%10.6f \n", dot_R3(ortho_b,ortho_b));
  };      

  gsl_rng_free (r);
  return 0;
};


int test_get_nearby_point(void){
  double glong_deg_in; /* sky coords gal long deg */
  double glat_deg_in;       /* sky coords gal lat deg */
  struct xyzw_s posn_in;
  double radius;
  double distance;     /* distance to new point */
  int    k_curv;         /* sign of curvature  */
  double R_C;  /* curvature radius if k_curv != 0 */
  double rSLS;           /* radius of SLS */
  double sin_rSLS;   /* rSLS orthogonally projected to R^3 = R_C sin(rSLS/R_C) */
  double glong_deg_out;  /* OUTPUTS from get_nearby_point: sky coords */
  double glat_deg_out;
  struct xyzw_s posn_out;
  double d_out;
  double vec_in[3], vec_out[3];

  /* declarations for GSL pseudo-random number generator */
  const gsl_rng_type *T;
  gsl_rng *r;

  int i, n=100;
  int i_get_nearby_point; /* return value */

  /* set up the generator */
  gsl_rng_env_setup();
  T = gsl_rng_default;
  r = gsl_rng_alloc (T);

  for(k_curv = 0; k_curv < 2; k_curv++){
    printf("test_get_nearby_point: k_curve = %d\n",k_curv);

    if(0==k_curv){
      R_C = 0; /* flat test */
    }else if(1==k_curv){
      R_C = 30000.0; /* example test value for rSLS < 10000 Mpc */
    };

    
    for(i=0;i<n;i++){
      rSLS = gsl_rng_uniform(r) * 10000.0;
      distance = (0.01 + gsl_rng_uniform(r)*0.98)*rSLS;
      /* distance = 1.999*rSLS; */
      glat_deg_in = 180.0 * gsl_rng_uniform (r) - 90.0;
      glong_deg_in = 360.0 * gsl_rng_uniform (r);

      printf("test_get_nearby_point: lb in = %.2f %.2f,  ",
	     glong_deg_in,glat_deg_in);
      printf("distance = %.2f  rSLS = %.2f \n",distance,rSLS);

      /* choose a point about \pm 5% inside/outside SLS */
      radius = rSLS * (1.0 + (0.1*gsl_rng_uniform(r))-0.05);
      posn_in.x = radius*cos(glat_deg_in*M_PI/180.0) 
	*cos(glong_deg_in*M_PI/180.0) ;
      posn_in.y = radius*cos(glat_deg_in*M_PI/180.0) 
	*sin(glong_deg_in*M_PI/180.0) ;
      posn_in.z = radius*sin(glat_deg_in*M_PI/180.0);

      if(0==k_curv){
	posn_in.w = 1.0;  /* not used */
      }else if(1==k_curv){
	posn_in.w = sqrt(R_C*R_C - posn_in.x*posn_in.x 
			 - posn_in.y*posn_in.y
			 - posn_in.z*posn_in.z);
      };

      i_get_nearby_point = get_nearby_point(posn_in,
					    distance,     
					    k_curv,      
					    R_C,
					    rSLS,
					    0,
					    &glong_deg_out,
					    &glat_deg_out
					    );

      printf("test_get_nearby_point: lb out = %.2f %.2f,  ",
	     glong_deg_out,glat_deg_out);

      if(0==k_curv){
	vec_in[0] = posn_in.x;
	vec_in[1] = posn_in.y;
	vec_in[2] = posn_in.z;

	vec_out[0] = rSLS* cos(glat_deg_out*M_PI/180.0) *  cos(glong_deg_out*M_PI/180.0);
	vec_out[1] = rSLS* cos(glat_deg_out*M_PI/180.0) *  sin(glong_deg_out*M_PI/180.0);
	vec_out[2] = rSLS* sin(glat_deg_out*M_PI/180.0);

	/* d_out = acos(dot_R3(vec_in,vec_out)) * rSLS; */
	d_out = dist_R3(vec_in,vec_out);

	/* test cos and acos as inverses */
	/* d_out = acos( cos(distance/rSLS) ) *rSLS; 
	 */
      }else if(1==k_curv){
	sin_rSLS = R_C *sin(rSLS/R_C) ;
	posn_out.x = sin_rSLS* cos(glat_deg_out*M_PI/180.0) *  cos(glong_deg_out*M_PI/180.0);
	posn_out.y = sin_rSLS* cos(glat_deg_out*M_PI/180.0) *  sin(glong_deg_out*M_PI/180.0);
	posn_out.z = sin_rSLS* sin(glat_deg_out*M_PI/180.0);
	posn_out.w = R_C *cos(rSLS/R_C) ;
	d_out = R_C * c_angle_R4(posn_in, posn_out, 4);
      };

      if(0==i_get_nearby_point){
	printf("d_out = %.2f\n",d_out);
	printf("dd: %.4f %.4f %.15f %.15f\n", 
	       glong_deg_in,glat_deg_in,
	       distance/rSLS, d_out/rSLS);
	printf("test_get_nearby_point: TEST_1 distance - d_out should be small: %10g\n",
	       distance - d_out);
      }else if(-1==i_get_nearby_point){
	printf("test_get_nearby_point: no intersection TEST_1 -99e9\n");
      }else{
	printf("test_get_nearby_point: failed with return value %d\n",
	       i_get_nearby_point);
      };
      printf("\n");

    }; /* for(i=0;i<n;i++) */
  };  /* for(k_curv = 0; k_curv < 2; k_curv++) */
  return 0;
};


int test_rotate_about_axis3D(void){
  double glong_axis; /* INPUTS: axis gal long radians */
  double glat_axis;       /* sky coords gal lat radians */
  double theta;          /* amount to rotate radians */
  double glong_in;       /* position to rotate g long */
  double glat_in;
  double lb_axis[3],ortho_a[3],ortho_b[3];
  double vec_in[3];
  double dummy[3];

  /* declarations for GSL pseudo-random number generator */
  const gsl_rng_type *T;
  gsl_rng *r;

  double d0,d1,d2;  /* dot products - projections onto orthonormal axes */
  double d0_new, d1_new,d2_new; /* after rotation */
  double radius;
  int i,j, n=4;

  /* set up the generator */
  gsl_rng_env_setup();
  T = gsl_rng_default;
  r = gsl_rng_alloc (T);

  for(i=0;i<n;i++){
    theta = M_PI * gsl_rng_uniform(r);
    glat_axis = M_PI * gsl_rng_uniform (r) - M_PI_2;
    glong_axis = 2.0*M_PI * gsl_rng_uniform (r);

    printf("test_rotate_about_axis3D: lb axis = %.2f %.2f,  ",
	   glong_axis,glat_axis);
    printf("theta = %.2f\n\n",theta);
    
    /* axis basis */
    get_orth_axes(glong_axis,glat_axis, lb_axis, ortho_a, ortho_b);

    for(j=0;j<n+3;j++){
      glat_in = M_PI * gsl_rng_uniform (r) - M_PI_2;
      glong_in = 2.0*M_PI * gsl_rng_uniform (r);
      get_orth_axes(glong_in,glat_in, vec_in, dummy,dummy);

      /* see what happens when rotation the orthonormal axes */
      if(j==n){
	vec_in[0]=lb_axis[0];
	vec_in[1]=lb_axis[1];
	vec_in[2]=lb_axis[2];
      }else if(j==n+1){
	vec_in[0]=ortho_a[0];
	vec_in[1]=ortho_a[1];
	vec_in[2]=ortho_a[2];
      }else if(j==n+2){
	vec_in[0]=ortho_b[0];
	vec_in[1]=ortho_b[1];
	vec_in[2]=ortho_b[2];
      };
      if(j>=n)c_xyz_rlb(vec_in[0], vec_in[1], vec_in[2], 
		  &radius, &glong_in, &glat_in);


      d0 = dot_R3(vec_in,lb_axis);
      d1 = dot_R3(vec_in,ortho_a);
      d2 = dot_R3(vec_in,ortho_b);

      printf("j= %d test_rotate_about_axis3D:\n before rotation:     lb in = %.2f %.2f  ",
	     j,glong_in,glat_in);
      printf("in axis basis vec_in = %10.3f %10.3f %10.3f, cos(theta)=%10.4f\n",
	     d0,d1,d2, cos(theta));

      /* should deg -> deg work here ? No. */
      /* glong_in *= 180.0/M_PI;
	 glat_in  *= 180.0/M_PI; */

      rotate_about_axis3D(glong_axis,glat_axis, theta, 
			  &glong_in, &glat_in);


      get_orth_axes(glong_in,glat_in, vec_in, dummy,dummy); 

      /* this will fail:
      get_orth_axes(glong_in*M_PI/180.0,
		    glat_in*M_PI/180.0, 
		    vec_in, dummy,dummy);
      */

      d0_new = dot_R3(vec_in,lb_axis);
      d1_new = dot_R3(vec_in,ortho_a);
      d2_new = dot_R3(vec_in,ortho_b);
      printf("after rotation: j, lb in = %d  %.2f %.2f,  ",
	     j,glong_in,glat_in);
      printf("in axis basis vec_in = %10.3f %10.3f %10.3f, cos(theta)=%10.4f\n",
	     d0_new,d1_new,d2_new, cos(theta));
      
    };
    printf("\n");
  };
  return 0;
};

      
      

