/* 
   astromisc - library with miscellaneous GPL tools for astronomy research
   genholon.h - header file for defining holonomies of the Poincare space

   Copyright (C) 2004 Boud Roukema

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
*/


/* holonomy prototypes which may need to be called externally to the
   astromisc package */
/* #ifndef ASTROMISC_HOLONOMIES */

#ifndef GENHOLON_FN
#define GENHOLON_FN 1

/* maximum order of a space, and order of individual spaces */
#ifndef MAX_ORDER
#define MAX_ORDER 120 
#endif

/* maximum number of axes for defining a holonomy */
#ifndef MAX_AXES
#define MAX_AXES 3
#endif

#ifndef XYZW_S
#define XYZW_S 1
/* following line obsolete: please replace by sizeof(struct xyzw_s) */
#define XYZW_STRUCT_SIZE 32
struct xyzw_s
{
  double x;
  double y;
  double z;
  double w;
};
#endif

/* calculate the MAX_ORDER holonomies as 4x4 matrices */
int genholon(int i_space,                            /* which space */ 
	     double holonomies_R4[MAX_ORDER][4][4],
             double gal_l, /* face centre if dodecahedral space */
             double gal_b,
             double theta, /* third orientation parameter if dodecspace */
             int    handedness,  /* -1 = lefthanded, +1 = righthanded */
             int    allow_arb_twist, /* 0 = don't allow, 1 = allow */
             double arb_twist  /* arbitrary twist angle */
);

struct holon_coords_s{
  int n_axes;
  double gal_l[MAX_AXES]; /* gal longitude of axes */
  double gal_b[MAX_AXES]; /* gal latitude of axes */
  double L_axis[MAX_AXES]; /* comoving lengths of axes */
};



/* calculate the MAX_ORDER holonomies as 4x4 matrices */
int genholon_flat(int i_space,                            /* which space */ 
		  double holonomies_R4[MAX_ORDER][4][4],
		  struct holon_coords_s holon_coords
		  ); 

int check_genholon_flat(void);


/* OUTPUTS: alpha_(minus|plus) are in radians */
int annulus_limits(int k_curv,  /* INPUTS: normalised sign of curvature */
		   double R_C,  /* curvature radius if k_curv != 0 */
		   double r_SLS, /* radius of SLS */
		   double d_inj, /* twice the injectivity radius */
		   double r_2,   /* maximum cross-separation */
		   double *alpha_minus, /*OUTPUTS: min angle from face centre */
		   double *alpha_plus); /*OUTPUTS: max angle from face centre */


/* Given a 4D position, the radius to the surface of last scattering,
   and a spatial geodesic distance, randomly choose another point on
   the SLS at that distance, from a uniform distribution along the
   circle of possible points.  TODO: Add the non-flat cases.  */
int get_nearby_point(struct xyzw_s posn_in, /* INPUTS: spatial position */
		     double distance,     /* distance to new point */
		     int    k_curv,         /* sign of curvature  */
		     double R_C,  /* curvature radius if k_curv != 0 */
		     double rSLS,           /* radius of SLS */
		     int    reset_gsl_rng,   /* free up gsl rng after usage? */
		     double *glong_deg_out,  /* OUTPUTS: sky coords */
		     double *glat_deg_out
		     );


#define SINCOS_MAX_TOL 1e-11
#define SMALL_R_TOL  1e-11

/*c_angle_R4 - angle between two vectors in R4, possibly using only some dims*/
/* return value in radians */
double c_angle_R4(struct xyzw_s vec_a, /* INPUT vectors */
		  struct xyzw_s vec_b, 
		  int n_dim            /* use just 3 or all 4 dims? */
		  );

/*c_hyperangle_R4 - hyperangle between two vectors in R4, possibly using only some dims*/
/* return value in radians */
double c_hyperangle_R4(struct xyzw_s vec_a, /* INPUT vectors */
		       struct xyzw_s vec_b, 
		       int n_dim            /* use just 3 or all 4 dims? */
		       );


double separation_3D(int k_curv,  /* INPUT: curvature sign*/
		     double R_C,   /*  radius of curvature  */
		     struct xyzw_s vec_a,  /* input vectors in 4D embedding if k_curv!=0*/
		     struct xyzw_s vec_b
		     );

double separation_3D_origin_w(int k_curv,  /* INPUT: curvature sign*/
			      double R_C,   /*  radius of curvature  */
			      struct xyzw_s vec_a  /* input vectors in 4D embedding if k_curv!=0*/
			      );

#endif
/* endif: #ifndef GENHOLON_FN */

#ifndef CORRNFN_DEFS
#define CORRNFN_DEFS 1
struct xyzw_s sim_galaxy_second_point(int k_curv, /* INPUTS */
				      double R_C,
				      struct xyzw_s first_galaxy,
				      double r_corrn_0,  /* Mpc */
				      double gamma_corrn,  /* e.g. 1.8 */
				      double r_min,
				      double r_max
				      );

int sim_galaxy_second_point_gsl_free(void);

#endif


int space_order(int i_space);

/* twist angle (first layer) is M_PI/space_ntwist if Clifford translation */
int space_ntwist(int i_space);

int n_spaces(void); /* returns N_SPACES */

const char *space_name(int i_space); /* returns name of  (i_space)-th space  */

#ifndef C_XYZ_RLB_FN
#define C_XYZ_RLB_FN 1
int c_xyz_rlb(double x_in,  /* inputs */
	      double y_in,
	      double z_in, 
	      double *radius, /* outputs */
	      double *glong,
	      double *glat);
#endif

/* Given a rotational axis sky position (glong_axis,glat_axis), a sky
   position to be rotated (glong_in,glat_in) and a rotation angle
   (radians), rotate the g*_in position by the rotation angle. */
   
int rotate_about_axis3D(double glong_axis, /* INPUTS: axis gal long radians */
			double glat_axis,       /* sky coords gal lat radians */
			double theta,          /* amount to rotate radians */
			double *glong_in,       /* position to rotate g long */
			double *glat_in);       /*           gal lat radians */

int test_rotate_about_axis3D(void);
