/* 
   astromisc - library with miscellaneous GPL tools for astronomy research
   separation_3D - metric separation of two points in 2D/3D constant curvature space

   Copyright (C) 2011-2012 Boud Roukema

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.

   See also http://www.gnu.org/licenses/gpl.html
*/

#include <stdio.h>
#include <strings.h>
#include <math.h>
#include <stdlib.h>
#include "holonomies.h"

/*
  Calculate the metric separation of two points in a constant curvature
  3D space of any of -, 0, or + curvature; embed in (n+1)D space 
  for the non-zero curvature cases. Numerical tolerances for division by
  zero are set by the parameters *_TOL in holonomies.h.
 */
double separation_3D(int k_curv, double R_C, struct xyzw_s vec_a, struct xyzw_s vec_b)
{
  double xx,yy,zz,r1sq, separation;

  int debug=0;

  if(0==k_curv){
    xx=vec_a.x-vec_b.x;
    yy=vec_a.y-vec_b.y;
    zz=vec_a.z-vec_b.z;
    r1sq =  xx*xx + yy*yy + zz*zz;
    if(r1sq < 0.0)
      printf("separation_3D software ERROR: r1sq = %g < 0.0\n",r1sq);
    separation= sqrt(r1sq);
  }else if(1==k_curv){
    if(debug)printf("separation_3D DEBUG: k=1\n");
    separation = R_C * c_angle_R4(vec_a, vec_b, 4);
  }else if(-1==k_curv){
    separation = R_C * c_hyperangle_R4(vec_a, vec_b, 4);
  }else{
    printf("separation_3D ERROR: k_curv = %d invalid.\n",k_curv);
    exit(EXIT_FAILURE);
  };
  return separation;
}


/*
  Calculate the metric distance to a point in a constant curvature
  3D space of any of -, 0, or + curvature from the origin
  which is at:
  * (0,0,0,w=R_C) for 3D curved spaces;
  * (0,0,0,1) for 3D flat space;  [the w=1 is not used]
  Each case is embedded in a (3+1)D Euclidean space with an appropriate
  intrinsic metric.
 */
double separation_3D_origin_w(int k_curv, double R_C, struct xyzw_s vec_a)
{
  double separation;
  struct xyzw_s xyzw_0;

  /* set the origin */
  xyzw_0.x = 0.0;
  xyzw_0.y = 0.0;
  xyzw_0.z = 0.0;
  if(0==k_curv){
    xyzw_0.w = 1.0; /* not needed: avoid calculating (1-1)^2 */
  }else{
    xyzw_0.w = R_C;
  };

  separation = separation_3D(k_curv, R_C, vec_a, xyzw_0);
  return separation;
};


int test_separation_3D(void){
  struct xyzw_s vec_a;
  struct xyzw_s vec_b;
  double R_C = 6.789;
  int j,k;

  double dist_projected= 0.12345;
  double out_dist;
  double out_dist_projected=-9e9;

  vec_a.x = 0.0;
  vec_a.y = 0.0;
  vec_a.z = 0.0;
  /* second point at projected distance = dist_projected  */
  vec_b.x = vec_a.x + dist_projected /sqrt(3);
  vec_b.y = vec_a.y - dist_projected /sqrt(3);
  vec_b.z = vec_a.z + dist_projected /sqrt(3);

  for(j=0;j<2;j++){
    for(k=-1;k<2;k++){
      vec_a.w = sqrt( R_C*R_C -
		      k * (vec_a.x*vec_a.x + vec_a.y*vec_a.y + vec_a.z*vec_a.z));
      vec_b.w = sqrt( R_C*R_C - 
		      k * (vec_b.x*vec_b.x + vec_b.y*vec_b.y + vec_b.z*vec_b.z));
      if(0==k){
	vec_a.w =1.0;
	vec_b.w =1.0;
      };
      if(0==j){
	out_dist = separation_3D(k, R_C, vec_a, vec_b); 
	printf("separation_3D(): \n");
      }else{
	out_dist = separation_3D_origin_w(k, R_C, vec_b); 
	printf("separation_3D_origin_w(): \n");
      };
      if(0==k){
	printf("k=0\n");
	out_dist_projected=out_dist;
      }else if(-1==k){
	printf("k=-1\n");
	out_dist_projected=R_C*sinh(out_dist/R_C);
      }else if(1==k){
	printf("k=1\n");
	out_dist_projected=R_C*sin(out_dist/R_C);
      };

      printf("test_separation_3D: input dist_projected= %20.15f, out_dist = %20.15f\n",
	     dist_projected,out_dist);
      printf("out_dist implies projected dist of %20.15f\n",
	     out_dist_projected);
    }; /* for(k=-1;k<2;k++) */
    printf("\n");
  };  /* for(j=0;j<2;j++) */



  return 0;
};
