/* 
   sim_galaxy_second_point - given a random galaxy, simulate a second galaxy
   astromisc - library with miscellaneous GPL tools for astronomy research

   Copyright (C) 2004 Boud Roukema

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2, or (at your option)
   any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software Foundation,
   Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.  
*/


#include "astromisc.h"
#include "holonomies.h"

/* global variables within this file */
const gsl_rng_type * T_gsl;
static gsl_rng * r_gsl;
static int gsl_first_call = 1;

#define MAX_TRY 1000000


/* For a randomly chosen galaxy at given XYZW position and a given
   two-point galaxy auto-correlation function 
     xi(r) =  (r/r_corrn_0)^gamma_corrn,
   choose a galaxy in a random direction at a distance
   of probability proportional to  
     (1+xi(r)) dr.
   The random choice of a galaxy must be done prior to
   calling this function. The direction is chosenly randomly
   over the 2-sphere.

   TODO: the non-flat cases - do at origin (0,0,0,R_C) and
   rotate into place.
*/
struct xyzw_s sim_galaxy_second_point(int k_curv, /* INPUTS */
				      double R_C,
				      struct xyzw_s first_galaxy,
				      double r_corrn_0,  /* Mpc */
				      double gamma_corrn,  /* e.g. 1.8 */
				      double r_min,
				      double r_max
				      ){

  int    i_try=0;
  struct xyzw_s second_galaxy;
  double xi_max;
  double r_try, xi_try, xi_test; /* generate pairs and test */

  int debug=0;

  if(1==gsl_first_call){
    gsl_first_call = 0;
    gsl_rng_env_setup();
    T_gsl= gsl_rng_default;
    gsl_rng_default_seed += 6977; /* add a prime to differ from other calls */
    r_gsl= gsl_rng_alloc(T_gsl);
  };

  xi_max = 1.0 + exp( log(r_min/r_corrn_0) * (-gamma_corrn) );

  if(0!=k_curv){
    printf("sim_galaxy_second_point: Error: non-curved cases not yet programmed.\n");
    exit(EXIT_FAILURE);
  }else{
    /* flat case */
    for(i_try=0; 
	i_try<MAX_TRY && (0==i_try || xi_test > xi_try);
	i_try++){
      second_galaxy.x = first_galaxy.x + gsl_rng_uniform(r_gsl) *2.0*r_max - r_max;
      second_galaxy.y = first_galaxy.y + gsl_rng_uniform(r_gsl) *2.0*r_max - r_max;
      second_galaxy.z = first_galaxy.z + gsl_rng_uniform(r_gsl) *2.0*r_max - r_max;
      second_galaxy.w = 1.0;
      r_try = separation_3D(k_curv, R_C, first_galaxy, second_galaxy);
      if(r_try > r_min){
	xi_try = 1.0 + exp( log(r_try/r_corrn_0) * (-gamma_corrn) );
      }else{
	xi_try = xi_max;
      };
      xi_test = gsl_rng_uniform(r_gsl) * xi_max;
    };
    if(i_try >= MAX_TRY){
      printf("sim_galaxy_second_point: WARNING: i_try = %d > %d\n",
	     i_try,MAX_TRY);
      printf("  r_corrn_0 = %f, gamma_corrn = %f, r_min = %f, r_max = %f, r_try = %f\n",
	     r_corrn_0, gamma_corrn, r_min, r_max, r_try);
    };
    if(debug){
      printf("sim_galaxy_second_point: OK: i_try = %d\n",i_try);
    };
  };
  return second_galaxy;
}


#define N_GAL_TEST 5000
#define MAX_TRY_TEST 100
int test_sim_galaxy_second_point(void){

  int k_curv = 0;
  double R_C = 1.0;
  struct xyzw_s galaxy[N_GAL_TEST];
  double r_corrn_0 = 5.0;  /* Mpc */
  double gamma_corrn = 1.8;  /* e.g. 1.8 */
  double r_min = 0.3;
  double r_max = 50.0; 
  double box_size = 3000.0;
  double uniform_percent = 0.5; /* = 0.0001; */ /* 0.002; */
  double volume; /* 0.002; */
  int    n_uniform;

  int i_gal, j_gal, n_gal;
  int i_try = 0;
  struct xyzw_s xyzw_try;
  int found_second;

  int debug = 1;

  if(1==gsl_first_call){
    gsl_first_call = 0;
    gsl_rng_env_setup();
    T_gsl= gsl_rng_default;
    gsl_rng_default_seed += 8573; /* add a prime to differ from other calls */
    r_gsl= gsl_rng_alloc(T_gsl);
  };

  volume = box_size*box_size*box_size;
  
  /*
  uniform_percent = fmin(1.0, (volume / (4.1888 * r_max*r_max*r_max))/(double)N_GAL_TEST ); 
  */
  /* 4pi/3 = 4.1888... */

  n_uniform=(int)(fmax(1.01,fmin(N_GAL_TEST+0.01,
				 uniform_percent*N_GAL_TEST)));

  if(debug){
    printf("uniform_percent = %f\n",uniform_percent);
    printf("n_uniform = %d\n",n_uniform );
  };

  /* choose a few percent of the points from a uniform distribution */
  for(i_gal=0; i_gal < n_uniform; i_gal++){
    if(debug)printf("i_gal=%d.....",i_gal);  
    galaxy[i_gal].x= gsl_rng_uniform(r_gsl) *box_size;
    galaxy[i_gal].y= gsl_rng_uniform(r_gsl) *box_size;
    galaxy[i_gal].z= gsl_rng_uniform(r_gsl) *box_size;
    galaxy[i_gal].w=1.0;

    printf("test_sim_galaxy_second: xyz_unif: %f %f %f %f \n",
	   galaxy[i_gal].x, galaxy[i_gal].y, galaxy[i_gal].z, galaxy[i_gal].w);
  };
  n_gal=i_gal;

  /* Successively generate each i_gal-th galaxy in relation to any of
     the points created earlier, until enough are obtained in the
     desired region. */
  for(i_gal=n_gal+0; i_gal < N_GAL_TEST; i_gal++){

    /* arbitrarily choose an existing galay */
    j_gal = (gsl_rng_uniform(r_gsl)*(double)(i_gal-1));
    if(debug)printf("  correlated points: i_gal, j_gal = %d %d\n",i_gal,j_gal);

    if(j_gal < 0)j_gal = 0;
    if(j_gal >= i_gal)j_gal = i_gal-1; /* the i_gal-th galaxy does not yet exist */

    found_second= 0;
    for(i_try=0; 
	i_try < MAX_TRY_TEST && 0==found_second;
	i_try++){
      xyzw_try = sim_galaxy_second_point( k_curv, /* INPUTS */
					    R_C,
					    galaxy[j_gal],
					    r_corrn_0,  /* Mpc */
					    gamma_corrn,  /* e.g. 1.8 */
					    r_min,
					    r_max);
      if( 0.0 <= xyzw_try.x && xyzw_try.x <= box_size &&
	  0.0 <= xyzw_try.y && xyzw_try.y <= box_size &&
	  0.0 <= xyzw_try.z && xyzw_try.z <= box_size ){
	found_second = 1;
      };
    }; /*     for(i_try=0 ... ) */
    printf("test_sim_galaxy_second: i_try= %d, found_second = %d\n",
	   i_try, found_second);
    if(found_second){
      printf("test_sim_galaxy_second: xyz_corr: %f %f %f %f \n",
	     xyzw_try.x, xyzw_try.y, xyzw_try.z, xyzw_try.w);
      galaxy[i_gal] = xyzw_try; /* copy the tried structure to the
				       list of structures */
    };
  }; /*   for(i_gal=n_gal-1; i_gal < N_GAL_TEST-1; i_gal++) */   


  sim_galaxy_second_point_gsl_free();
  return 0;
}



int sim_galaxy_second_point_gsl_free(void){
  if(0==gsl_first_call){
    gsl_first_call = 1;
    gsl_rng_free(r_gsl);
    return 0;
  }else{ /* nothing to be freed */
    return 1;
  };

}
